@@ -5,8 +5,8 @@ LZ4 Decompress
 
 #ifdef HAVE_LIBLZ4
 
-#include <stdio.h>
 #include <lz4frame.h>
+#include <stdio.h>
 
 #include "common/compress/lz4/common.h"
 #include "common/compress/lz4/decompress.h"
@@ -32,18 +32,18 @@ typedef struct Lz4Decompress
 /***********************************************************************************************************************************
 Render as string for logging
 ***********************************************************************************************************************************/
-static String *
-lz4DecompressToLog(const Lz4Decompress *this)
+static void
+lz4DecompressToLog(const Lz4Decompress *const this, StringStatic *const debugLog)
 {
-    return strNewFmt(
-        "{inputSame: %s, inputOffset: %zu, frameDone %s, done: %s}", cvtBoolToConstZ(this->inputSame), this->inputOffset,
-        cvtBoolToConstZ(this->frameDone), cvtBoolToConstZ(this->done));
+    strStcFmt(
+        debugLog, "{inputSame: %s, inputOffset: %zu, frameDone %s, done: %s}", cvtBoolToConstZ(this->inputSame),
+        this->inputOffset, cvtBoolToConstZ(this->frameDone), cvtBoolToConstZ(this->done));
 }
 
 #define FUNCTION_LOG_LZ4_DECOMPRESS_TYPE                                                                                           \
     Lz4Decompress *
 #define FUNCTION_LOG_LZ4_DECOMPRESS_FORMAT(value, buffer, bufferSize)                                                              \
-    FUNCTION_LOG_STRING_OBJECT_FORMAT(value, lz4DecompressToLog, buffer, bufferSize)
+    FUNCTION_LOG_OBJECT_FORMAT(value, lz4DecompressToLog, buffer, bufferSize)
 
 /***********************************************************************************************************************************
 Free decompression context
@@ -135,7 +135,7 @@ lz4DecompressDone(const THIS_VOID)
 
     ASSERT(this != NULL);
 
-    FUNCTION_TEST_RETURN(this->done);
+    FUNCTION_TEST_RETURN(BOOL, this->done);
 }
 
 /***********************************************************************************************************************************
@@ -152,36 +152,34 @@ lz4DecompressInputSame(const THIS_VOID)
 
     ASSERT(this != NULL);
 
-    FUNCTION_TEST_RETURN(this->inputSame);
+    FUNCTION_TEST_RETURN(BOOL, this->inputSame);
 }
 
 /**********************************************************************************************************************************/
-IoFilter *
-lz4DecompressNew(void)
+FN_EXTERN IoFilter *
+lz4DecompressNew(const bool raw)
 {
-    FUNCTION_LOG_VOID(logLevelTrace);
+    FUNCTION_LOG_BEGIN(logLevelTrace);
+        (void)raw;                                                  // Not required for decompress
+    FUNCTION_LOG_END();
 
-    IoFilter *this = NULL;
-
-    OBJ_NEW_BEGIN(Lz4Decompress)
+    OBJ_NEW_BEGIN(Lz4Decompress, .childQty = MEM_CONTEXT_QTY_MAX, .callbackQty = 1)
     {
-        Lz4Decompress *driver = OBJ_NEW_ALLOC();
-        *driver = (Lz4Decompress){0};
+        *this = (Lz4Decompress){0};
 
         // Create lz4 context
-        lz4Error(LZ4F_createDecompressionContext(&driver->context, LZ4F_VERSION));
+        lz4Error(LZ4F_createDecompressionContext(&this->context, LZ4F_VERSION));
 
         // Set callback to ensure lz4 context is freed
-        memContextCallbackSet(objMemContext(driver), lz4DecompressFreeResource, driver);
-
-        // Create filter interface
-        this = ioFilterNewP(
-            LZ4_DECOMPRESS_FILTER_TYPE, driver, NULL, .done = lz4DecompressDone, .inOut = lz4DecompressProcess,
-            .inputSame = lz4DecompressInputSame);
+        memContextCallbackSet(objMemContext(this), lz4DecompressFreeResource, this);
     }
     OBJ_NEW_END();
 
-    FUNCTION_LOG_RETURN(IO_FILTER, this);
+    FUNCTION_LOG_RETURN(
+        IO_FILTER,
+        ioFilterNewP(
+            LZ4_DECOMPRESS_FILTER_TYPE, this, NULL, .done = lz4DecompressDone, .inOut = lz4DecompressProcess,
+            .inputSame = lz4DecompressInputSame));
 }
 
 #endif // HAVE_LIBLZ4
