@@ -6,8 +6,8 @@ GCS Storage
 #include <string.h>
 
 #include <openssl/bio.h>
-#include <openssl/x509.h>
 #include <openssl/pem.h>
+#include <openssl/x509.h>
 
 #include "common/crypto/common.h"
 #include "common/crypto/hash.h"
@@ -22,7 +22,6 @@ GCS Storage
 #include "common/type/json.h"
 #include "common/type/object.h"
 #include "storage/gcs/read.h"
-#include "storage/gcs/storage.intern.h"
 #include "storage/gcs/write.h"
 #include "storage/posix/storage.h"
 
@@ -53,19 +52,19 @@ VARIANT_STRDEF_STATIC(GCS_JSON_ERROR_VAR,                           "error");
 VARIANT_STRDEF_STATIC(GCS_JSON_ERROR_DESCRIPTION_VAR,               "error_description");
 VARIANT_STRDEF_STATIC(GCS_JSON_EXPIRES_IN_VAR,                      "expires_in");
 #define GCS_JSON_ITEMS                                              "items"
-    VARIANT_STRDEF_STATIC(GCS_JSON_ITEMS_VAR,                       GCS_JSON_ITEMS);
+VARIANT_STRDEF_STATIC(GCS_JSON_ITEMS_VAR,                           GCS_JSON_ITEMS);
 VARIANT_STRDEF_EXTERN(GCS_JSON_MD5_HASH_VAR,                        GCS_JSON_MD5_HASH);
 VARIANT_STRDEF_EXTERN(GCS_JSON_NAME_VAR,                            GCS_JSON_NAME);
 #define GCS_JSON_NEXT_PAGE_TOKEN                                    "nextPageToken"
-    VARIANT_STRDEF_STATIC(GCS_JSON_NEXT_PAGE_TOKEN_VAR,             GCS_JSON_NEXT_PAGE_TOKEN);
+VARIANT_STRDEF_STATIC(GCS_JSON_NEXT_PAGE_TOKEN_VAR,                 GCS_JSON_NEXT_PAGE_TOKEN);
 #define GCS_JSON_PREFIXES                                           "prefixes"
-    VARIANT_STRDEF_STATIC(GCS_JSON_PREFIXES_VAR,                    GCS_JSON_PREFIXES);
+VARIANT_STRDEF_STATIC(GCS_JSON_PREFIXES_VAR,                        GCS_JSON_PREFIXES);
 VARIANT_STRDEF_STATIC(GCS_JSON_PRIVATE_KEY_VAR,                     "private_key");
 VARIANT_STRDEF_EXTERN(GCS_JSON_SIZE_VAR,                            GCS_JSON_SIZE);
 VARIANT_STRDEF_STATIC(GCS_JSON_TOKEN_TYPE_VAR,                      "token_type");
 VARIANT_STRDEF_STATIC(GCS_JSON_TOKEN_URI_VAR,                       "token_uri");
 #define GCS_JSON_UPDATED                                            "updated"
-    VARIANT_STRDEF_STATIC(GCS_JSON_UPDATED_VAR,                     GCS_JSON_UPDATED);
+VARIANT_STRDEF_STATIC(GCS_JSON_UPDATED_VAR,                         GCS_JSON_UPDATED);
 
 // Fields required when listing files
 #define GCS_FIELD_LIST                                                                                                             \
@@ -88,10 +87,10 @@ struct StorageGcs
     const String *bucket;                                           // Bucket to store data in
     const String *endpoint;                                         // Endpoint
     size_t chunkSize;                                               // Block size for resumable upload
+    const Buffer *tag;                                              // Tags to be applied to objects
 
     StorageGcsKeyType keyType;                                      // Auth key type
-    const String *credential;                                       // Credential (client email)
-    const String *privateKey;                                       // Private key in PEM format
+    const String *key;                                              // Key (value depends on key type)
     String *token;                                                  // Token
     time_t tokenTimeExpire;                                         // Token expiration time (if service auth)
     HttpUrl *authUrl;                                               // URL for authentication server
@@ -112,45 +111,51 @@ typedef struct
 } StorageGcsAuthTokenResult;
 
 static StorageGcsAuthTokenResult
-storageGcsAuthToken(HttpRequest *request, time_t timeBegin)
+storageGcsAuthToken(HttpRequest *const request, const time_t timeBegin)
 {
     FUNCTION_TEST_BEGIN();
         FUNCTION_TEST_PARAM(HTTP_REQUEST, request);
         FUNCTION_TEST_PARAM(TIME, timeBegin);
     FUNCTION_TEST_END();
 
+    FUNCTION_AUDIT_STRUCT();
+
     StorageGcsAuthTokenResult result = {0};
 
-    // Get the response
-    KeyValue *kvResponse = jsonToKv(strNewBuf(httpResponseContent(httpRequestResponse(request, true))));
-
-    // Check for an error
-    const String *error = varStr(kvGet(kvResponse, GCS_JSON_ERROR_VAR));
-
-    if (error != NULL)
+    MEM_CONTEXT_TEMP_BEGIN()
     {
-        THROW_FMT(
-            ProtocolError, "unable to get authentication token: [%s] %s", strZ(error),
-            strZNull(varStr(kvGet(kvResponse, GCS_JSON_ERROR_DESCRIPTION_VAR))));
+        // Get the response
+        KeyValue *const kvResponse = varKv(jsonToVar(strNewBuf(httpResponseContent(httpRequestResponse(request, true)))));
+
+        // Check for an error
+        const String *const error = varStr(kvGet(kvResponse, GCS_JSON_ERROR_VAR));
+
+        if (error != NULL)
+        {
+            THROW_FMT(
+                ProtocolError, "unable to get authentication token: [%s] %s", strZ(error),
+                strZNull(varStr(kvGet(kvResponse, GCS_JSON_ERROR_DESCRIPTION_VAR))));
+        }
+
+        MEM_CONTEXT_PRIOR_BEGIN()
+        {
+            // Get token
+            result.tokenType = strDup(varStr(kvGet(kvResponse, GCS_JSON_TOKEN_TYPE_VAR)));
+            CHECK(FormatError, result.tokenType != NULL, "token type missing");
+            result.token = strDup(varStr(kvGet(kvResponse, GCS_JSON_ACCESS_TOKEN_VAR)));
+            CHECK(FormatError, result.token != NULL, "access token missing");
+
+            // Get expiration
+            const Variant *const expiresIn = kvGet(kvResponse, GCS_JSON_EXPIRES_IN_VAR);
+            CHECK(FormatError, expiresIn != NULL, "expiry missing");
+
+            result.timeExpire = timeBegin + (time_t)varInt64Force(expiresIn);
+        }
+        MEM_CONTEXT_PRIOR_END();
     }
+    MEM_CONTEXT_TEMP_END();
 
-    MEM_CONTEXT_PRIOR_BEGIN()
-    {
-        // Get token
-        result.tokenType = strDup(varStr(kvGet(kvResponse, GCS_JSON_TOKEN_TYPE_VAR)));
-        CHECK(FormatError, result.tokenType != NULL, "token type missing");
-        result.token = strDup(varStr(kvGet(kvResponse, GCS_JSON_ACCESS_TOKEN_VAR)));
-        CHECK(FormatError, result.token != NULL, "access token missing");
-
-        // Get expiration
-        const Variant *const expiresIn = kvGet(kvResponse, GCS_JSON_EXPIRES_IN_VAR);
-        CHECK(FormatError, expiresIn != NULL, "expiry missing");
-
-        result.timeExpire = timeBegin + (time_t)varInt64Force(expiresIn);
-    }
-    MEM_CONTEXT_PRIOR_END();
-
-    FUNCTION_TEST_RETURN(result);
+    FUNCTION_TEST_RETURN_TYPE(StorageGcsAuthTokenResult, result);
 }
 
 /***********************************************************************************************************************************
@@ -172,14 +177,21 @@ storageGcsAuthJwt(StorageGcs *this, time_t timeBegin)
 
     MEM_CONTEXT_TEMP_BEGIN()
     {
+        // Load client email and private key
+        KeyValue *const kvKey = varKv(jsonToVar(strNewBuf(storageGetP(storageNewReadP(storagePosixNewP(FSLASH_STR), this->key)))));
+        const String *const clientEmail = varStr(kvGet(kvKey, GCS_JSON_CLIENT_EMAIL_VAR));
+        const String *const privateKeyRaw = varStr(kvGet(kvKey, GCS_JSON_PRIVATE_KEY_VAR));
+
+        CHECK(FormatError, clientEmail != NULL && privateKeyRaw != NULL, "credentials missing");
+
         // Add claim
         strCatEncode(
-            result, encodeBase64Url,
+            result, encodingBase64Url,
             BUFSTR(
                 strNewFmt(
                     "{\"iss\":\"%s\",\"scope\":\"https://www.googleapis.com/auth/devstorage.read%s\",\"aud\":\"%s\""
-                        ",\"exp\":%" PRIu64 ",\"iat\":%" PRIu64 "}",
-                    strZ(this->credential), this->write ? "_write" : "_only", strZ(httpUrl(this->authUrl)),
+                    ",\"exp\":%" PRIu64 ",\"iat\":%" PRIu64 "}",
+                    strZ(clientEmail), this->write ? "_write" : "_only", strZ(httpUrl(this->authUrl)),
                     (uint64_t)timeBegin + 3600, (uint64_t)timeBegin)));
 
         // Sign with RSA key
@@ -193,7 +205,7 @@ storageGcsAuthJwt(StorageGcs *this, time_t timeBegin)
         {
             // Load key
             bio = BIO_new(BIO_s_mem());
-            BIO_write((BIO *)bio, strZ(this->privateKey), (int)strSize(this->privateKey));
+            BIO_write((BIO *)bio, strZ(privateKeyRaw), (int)strSize(privateKeyRaw));
 
             privateKey = PEM_read_bio_PrivateKey((BIO *)bio, NULL, NULL, NULL);
             cryptoError(privateKey == NULL, "unable to read PEM");
@@ -216,7 +228,7 @@ storageGcsAuthJwt(StorageGcs *this, time_t timeBegin)
 
             // Add dot delimiter and signature
             strCatChr(result, '.');
-            strCatEncode(result, encodeBase64Url, signature);
+            strCatEncode(result, encodingBase64Url, signature);
         }
         FINALLY()
         {
@@ -232,7 +244,7 @@ storageGcsAuthJwt(StorageGcs *this, time_t timeBegin)
     }
     MEM_CONTEXT_TEMP_END();
 
-    FUNCTION_TEST_RETURN(result);
+    FUNCTION_TEST_RETURN(STRING, result);
 }
 
 static StorageGcsAuthTokenResult
@@ -243,6 +255,8 @@ storageGcsAuthService(StorageGcs *this, time_t timeBegin)
         FUNCTION_TEST_PARAM(TIME, timeBegin);
     FUNCTION_TEST_END();
 
+    FUNCTION_AUDIT_STRUCT();
+
     ASSERT(this != NULL);
     ASSERT(timeBegin > 0);
 
@@ -262,11 +276,15 @@ storageGcsAuthService(StorageGcs *this, time_t timeBegin)
         HttpRequest *request = httpRequestNewP(
             this->authClient, HTTP_VERB_POST_STR, httpUrlPath(this->authUrl), NULL, .header = header, .content = BUFSTR(content));
 
-        result = storageGcsAuthToken(request, timeBegin);
+        MEM_CONTEXT_PRIOR_BEGIN()
+        {
+            result = storageGcsAuthToken(request, timeBegin);
+        }
+        MEM_CONTEXT_PRIOR_END();
     }
     MEM_CONTEXT_TEMP_END();
 
-    FUNCTION_TEST_RETURN(result);
+    FUNCTION_TEST_RETURN_TYPE(StorageGcsAuthTokenResult, result);
 }
 
 /***********************************************************************************************************************************
@@ -282,6 +300,8 @@ storageGcsAuthAuto(StorageGcs *this, time_t timeBegin)
         FUNCTION_TEST_PARAM(TIME, timeBegin);
     FUNCTION_TEST_END();
 
+    FUNCTION_AUDIT_STRUCT();
+
     ASSERT(this != NULL);
     ASSERT(timeBegin > 0);
 
@@ -297,11 +317,15 @@ storageGcsAuthAuto(StorageGcs *this, time_t timeBegin)
         HttpRequest *request = httpRequestNewP(
             this->authClient, HTTP_VERB_GET_STR, httpUrlPath(this->authUrl), NULL, .header = header);
 
-        result = storageGcsAuthToken(request, timeBegin);
+        MEM_CONTEXT_PRIOR_BEGIN()
+        {
+            result = storageGcsAuthToken(request, timeBegin);
+        }
+        MEM_CONTEXT_PRIOR_END();
     }
     MEM_CONTEXT_TEMP_END();
 
-    FUNCTION_TEST_RETURN(result);
+    FUNCTION_TEST_RETURN_TYPE(StorageGcsAuthTokenResult, result);
 }
 
 /***********************************************************************************************************************************
@@ -312,7 +336,7 @@ storageGcsAuth(StorageGcs *this, HttpHeader *httpHeader)
 {
     FUNCTION_TEST_BEGIN();
         FUNCTION_TEST_PARAM(STORAGE_GCS, this);
-        FUNCTION_TEST_PARAM(KEY_VALUE, httpHeader);
+        FUNCTION_TEST_PARAM(HTTP_HEADER, httpHeader);
     FUNCTION_TEST_END();
 
     ASSERT(this != NULL);
@@ -331,10 +355,11 @@ storageGcsAuth(StorageGcs *this, HttpHeader *httpHeader)
             // If the current token has expired then request a new one
             if (timeBegin >= this->tokenTimeExpire)
             {
-                StorageGcsAuthTokenResult tokenResult = this->keyType == storageGcsKeyTypeAuto ?
-                    storageGcsAuthAuto(this, timeBegin) : storageGcsAuthService(this, timeBegin);
+                StorageGcsAuthTokenResult tokenResult =
+                    this->keyType == storageGcsKeyTypeAuto ?
+                        storageGcsAuthAuto(this, timeBegin) : storageGcsAuthService(this, timeBegin);
 
-                MEM_CONTEXT_BEGIN(THIS_MEM_CONTEXT())
+                MEM_CONTEXT_OBJ_BEGIN(this)
                 {
                     strFree(this->token);
                     this->token = strNewFmt("%s %s", strZ(tokenResult.tokenType), strZ(tokenResult.token));
@@ -343,7 +368,7 @@ storageGcsAuth(StorageGcs *this, HttpHeader *httpHeader)
                     this->tokenTimeExpire =
                         tokenResult.timeExpire - ((time_t)(httpClientTimeout(this->httpClient) / MSEC_PER_SEC * 2));
                 }
-                MEM_CONTEXT_END();
+                MEM_CONTEXT_OBJ_END();
             }
         }
 
@@ -358,7 +383,7 @@ storageGcsAuth(StorageGcs *this, HttpHeader *httpHeader)
 /***********************************************************************************************************************************
 Process Gcs request
 ***********************************************************************************************************************************/
-HttpRequest *
+FN_EXTERN HttpRequest *
 storageGcsRequestAsync(StorageGcs *this, const String *verb, StorageGcsRequestAsyncParam param)
 {
     FUNCTION_LOG_BEGIN(logLevelDebug);
@@ -367,6 +392,7 @@ storageGcsRequestAsync(StorageGcs *this, const String *verb, StorageGcsRequestAs
         FUNCTION_LOG_PARAM(BOOL, param.noBucket);
         FUNCTION_LOG_PARAM(BOOL, param.upload);
         FUNCTION_LOG_PARAM(BOOL, param.noAuth);
+        FUNCTION_LOG_PARAM(BOOL, param.tag);
         FUNCTION_LOG_PARAM(STRING, param.object);
         FUNCTION_LOG_PARAM(HTTP_HEADER, param.header);
         FUNCTION_LOG_PARAM(HTTP_QUERY, param.query);
@@ -390,9 +416,19 @@ storageGcsRequestAsync(StorageGcs *this, const String *verb, StorageGcsRequestAs
         if (param.object != NULL)
             strCatFmt(path, "/%s", strZ(httpUriEncode(strSub(param.object, 1), false)));
 
-        // Create header list and add content length
-        HttpHeader *requestHeader = param.header == NULL ?
-            httpHeaderNew(this->headerRedactList) : httpHeaderDup(param.header, this->headerRedactList);
+        // Create header list
+        HttpHeader *requestHeader =
+            param.header == NULL ? httpHeaderNew(this->headerRedactList) : httpHeaderDup(param.header, this->headerRedactList);
+
+        // Add tags
+        if (param.tag)
+        {
+            ASSERT(param.content == NULL);
+            ASSERT(this->tag != NULL);
+
+            httpHeaderPut(requestHeader, HTTP_HEADER_CONTENT_TYPE_STR, HTTP_HEADER_CONTENT_TYPE_JSON_STR);
+            param.content = this->tag;
+        }
 
         // Set host
         httpHeaderPut(requestHeader, HTTP_HEADER_HOST_STR, this->endpoint);
@@ -422,7 +458,7 @@ storageGcsRequestAsync(StorageGcs *this, const String *verb, StorageGcsRequestAs
     FUNCTION_LOG_RETURN(HTTP_REQUEST, result);
 }
 
-HttpResponse *
+FN_EXTERN HttpResponse *
 storageGcsResponse(HttpRequest *request, StorageGcsResponseParam param)
 {
     FUNCTION_LOG_BEGIN(logLevelDebug);
@@ -455,8 +491,8 @@ storageGcsResponse(HttpRequest *request, StorageGcsResponseParam param)
     FUNCTION_LOG_RETURN(HTTP_RESPONSE, result);
 }
 
-HttpResponse *
-storageGcsRequest(StorageGcs *this, const String *verb, StorageGcsRequestParam param)
+FN_EXTERN HttpResponse *
+storageGcsRequest(StorageGcs *const this, const String *const verb, const StorageGcsRequestParam param)
 {
     FUNCTION_LOG_BEGIN(logLevelDebug);
         FUNCTION_LOG_PARAM(STORAGE_GCS, this);
@@ -464,6 +500,7 @@ storageGcsRequest(StorageGcs *this, const String *verb, StorageGcsRequestParam p
         FUNCTION_LOG_PARAM(BOOL, param.noBucket);
         FUNCTION_LOG_PARAM(BOOL, param.upload);
         FUNCTION_LOG_PARAM(BOOL, param.noAuth);
+        FUNCTION_LOG_PARAM(BOOL, param.tag);
         FUNCTION_LOG_PARAM(STRING, param.object);
         FUNCTION_LOG_PARAM(HTTP_HEADER, param.header);
         FUNCTION_LOG_PARAM(HTTP_QUERY, param.query);
@@ -473,13 +510,15 @@ storageGcsRequest(StorageGcs *this, const String *verb, StorageGcsRequestParam p
         FUNCTION_LOG_PARAM(BOOL, param.contentIo);
     FUNCTION_LOG_END();
 
-    FUNCTION_LOG_RETURN(
-        HTTP_RESPONSE,
-        storageGcsResponseP(
-            storageGcsRequestAsyncP(
-                this, verb, .noBucket = param.noBucket, .upload = param.upload, .noAuth = param.noAuth, .object = param.object,
-                .header = param.header, .query = param.query, .content = param.content),
-            .allowMissing = param.allowMissing, .allowIncomplete = param.allowIncomplete, .contentIo = param.contentIo));
+    HttpRequest *const request = storageGcsRequestAsyncP(
+        this, verb, .noBucket = param.noBucket, .upload = param.upload, .noAuth = param.noAuth, .tag = param.tag,
+        .object = param.object, .header = param.header, .query = param.query, .content = param.content);
+    HttpResponse *const result = storageGcsResponseP(
+        request, .allowMissing = param.allowMissing, .allowIncomplete = param.allowIncomplete, .contentIo = param.contentIo);
+
+    httpRequestFree(request);
+
+    FUNCTION_LOG_RETURN(HTTP_RESPONSE, result);
 }
 
 /***********************************************************************************************************************************
@@ -488,17 +527,17 @@ General function for listing files to be used by other list routines
 // Helper to convert YYYY-MM-DDTHH:MM:SS.MSECZ format to time_t. This format is very nearly ISO-8601 except for the inclusion of
 // milliseconds, which are discarded here.
 static time_t
-storageGcsCvtTime(const String *time)
+storageGcsCvtTime(const String *const time)
 {
     FUNCTION_TEST_BEGIN();
         FUNCTION_TEST_PARAM(STRING, time);
     FUNCTION_TEST_END();
 
     FUNCTION_TEST_RETURN(
+        TIME,
         epochFromParts(
-            cvtZToInt(strZ(strSubN(time, 0, 4))), cvtZToInt(strZ(strSubN(time, 5, 2))),
-            cvtZToInt(strZ(strSubN(time, 8, 2))), cvtZToInt(strZ(strSubN(time, 11, 2))),
-            cvtZToInt(strZ(strSubN(time, 14, 2))), cvtZToInt(strZ(strSubN(time, 17, 2))), 0));
+            cvtZSubNToInt(strZ(time), 0, 4), cvtZSubNToInt(strZ(time), 5, 2), cvtZSubNToInt(strZ(time), 8, 2),
+            cvtZSubNToInt(strZ(time), 11, 2), cvtZSubNToInt(strZ(time), 14, 2), cvtZSubNToInt(strZ(time), 17, 2), 0));
 }
 
 static void
@@ -518,7 +557,7 @@ storageGcsInfoFile(StorageInfo *info, const KeyValue *file)
 static void
 storageGcsListInternal(
     StorageGcs *this, const String *path, StorageInfoLevel level, const String *expression, bool recurse,
-    StorageInfoListCallback callback, void *callbackData)
+    StorageListCallback callback, void *callbackData)
 {
     FUNCTION_LOG_BEGIN(logLevelDebug);
         FUNCTION_LOG_PARAM(STORAGE_GCS, this);
@@ -530,6 +569,8 @@ storageGcsListInternal(
         FUNCTION_LOG_PARAM_P(VOID, callbackData);
     FUNCTION_LOG_END();
 
+    FUNCTION_AUDIT_CALLBACK();
+
     ASSERT(this != NULL);
     ASSERT(path != NULL);
 
@@ -596,7 +637,7 @@ storageGcsListInternal(
                 else
                     response = storageGcsRequestP(this, HTTP_VERB_GET_STR, .query = query);
 
-                KeyValue *content = jsonToKv(strNewBuf(httpResponseContent(response)));
+                KeyValue *content = varKv(jsonToVar(strNewBuf(httpResponseContent(response))));
 
                 // If next page token exists then send an async request to get more data
                 const String *nextPageToken = varStr(kvGet(content, GCS_JSON_NEXT_PAGE_TOKEN_VAR));
@@ -687,7 +728,7 @@ storageGcsListInternal(
 
 /**********************************************************************************************************************************/
 static StorageInfo
-storageGcsInfo(THIS_VOID, const String *file, StorageInfoLevel level, StorageInterfaceInfoParam param)
+storageGcsInfo(THIS_VOID, const String *const file, const StorageInfoLevel level, const StorageInterfaceInfoParam param)
 {
     THIS(StorageGcs);
 
@@ -701,31 +742,53 @@ storageGcsInfo(THIS_VOID, const String *file, StorageInfoLevel level, StorageInt
     ASSERT(this != NULL);
     ASSERT(file != NULL);
 
-    // // Attempt to get file info
-    HttpResponse *httpResponse = storageGcsRequestP(
-        this, HTTP_VERB_GET_STR, .object = file, .allowMissing = true,
-        .query = httpQueryAdd(
-            httpQueryNewP(), GCS_QUERY_FIELDS_STR,
-            level >= storageInfoLevelBasic ? STRDEF(GCS_JSON_SIZE "," GCS_JSON_UPDATED) : EMPTY_STR));
+    StorageInfo result = {.level = level};
 
-    // Does the file exist?
-    StorageInfo result = {.level = level, .exists = httpResponseCodeOk(httpResponse)};
-
-    // Add basic level info if requested and the file exists
-    if (result.level >= storageInfoLevelBasic && result.exists)
+    MEM_CONTEXT_TEMP_BEGIN()
     {
-        result.type = storageTypeFile;
-        storageGcsInfoFile(&result, jsonToKv(strNewBuf(httpResponseContent(httpResponse))));
+        // Attempt to get file info
+        HttpResponse *const httpResponse = storageGcsRequestP(
+            this, HTTP_VERB_GET_STR, .object = file, .allowMissing = true,
+            .query = httpQueryAdd(
+                httpQueryNewP(), GCS_QUERY_FIELDS_STR,
+                level >= storageInfoLevelBasic ? STRDEF(GCS_JSON_SIZE "," GCS_JSON_UPDATED) : EMPTY_STR));
+
+        // Does the file exist?
+        result.exists = httpResponseCodeOk(httpResponse);
+
+        // Add basic level info if requested and the file exists
+        if (result.level >= storageInfoLevelBasic && result.exists)
+        {
+            result.type = storageTypeFile;
+            storageGcsInfoFile(&result, varKv(jsonToVar(strNewBuf(httpResponseContent(httpResponse)))));
+        }
+
+        httpResponseFree(httpResponse);
     }
+    MEM_CONTEXT_TEMP_END();
 
     FUNCTION_LOG_RETURN(STORAGE_INFO, result);
 }
 
 /**********************************************************************************************************************************/
-static bool
-storageGcsInfoList(
-    THIS_VOID, const String *path, StorageInfoLevel level, StorageInfoListCallback callback, void *callbackData,
-    StorageInterfaceInfoListParam param)
+static void
+storageGcsListCallback(void *const callbackData, const StorageInfo *const info)
+{
+    FUNCTION_TEST_BEGIN();
+        FUNCTION_TEST_PARAM_P(VOID, callbackData);
+        FUNCTION_TEST_PARAM(STORAGE_INFO, info);
+    FUNCTION_TEST_END();
+
+    ASSERT(callbackData != NULL);
+    ASSERT(info != NULL);
+
+    storageLstAdd(callbackData, info);
+
+    FUNCTION_TEST_RETURN_VOID();
+}
+
+static StorageList *
+storageGcsList(THIS_VOID, const String *const path, const StorageInfoLevel level, const StorageInterfaceListParam param)
 {
     THIS(StorageGcs);
 
@@ -733,22 +796,17 @@ storageGcsInfoList(
         FUNCTION_LOG_PARAM(STORAGE_GCS, this);
         FUNCTION_LOG_PARAM(STRING, path);
         FUNCTION_LOG_PARAM(ENUM, level);
-        FUNCTION_LOG_PARAM(FUNCTIONP, callback);
-        FUNCTION_LOG_PARAM_P(VOID, callbackData);
         FUNCTION_LOG_PARAM(STRING, param.expression);
     FUNCTION_LOG_END();
 
     ASSERT(this != NULL);
     ASSERT(path != NULL);
-    ASSERT(callback != NULL);
 
-    MEM_CONTEXT_TEMP_BEGIN()
-    {
-        storageGcsListInternal(this, path, level, param.expression, false, callback, callbackData);
-    }
-    MEM_CONTEXT_TEMP_END();
+    StorageList *const result = storageLstNew(level);
 
-    FUNCTION_LOG_RETURN(BOOL, true);
+    storageGcsListInternal(this, path, level, param.expression, false, storageGcsListCallback, result);
+
+    FUNCTION_LOG_RETURN(STORAGE_LIST, result);
 }
 
 /**********************************************************************************************************************************/
@@ -786,11 +844,12 @@ storageGcsNewWrite(THIS_VOID, const String *file, StorageInterfaceNewWriteParam
     ASSERT(this != NULL);
     ASSERT(file != NULL);
     ASSERT(param.createPath);
+    ASSERT(param.truncate);
     ASSERT(param.user == NULL);
     ASSERT(param.group == NULL);
     ASSERT(param.timeModified == 0);
 
-    FUNCTION_LOG_RETURN(STORAGE_WRITE, storageWriteGcsNew(this, file, this->chunkSize));
+    FUNCTION_LOG_RETURN(STORAGE_WRITE, storageWriteGcsNew(this, file, this->chunkSize, this->tag != NULL));
 }
 
 /**********************************************************************************************************************************/
@@ -803,7 +862,7 @@ typedef struct StorageGcsPathRemoveData
 } StorageGcsPathRemoveData;
 
 static void
-storageGcsPathRemoveCallback(void *callbackData, const StorageInfo *info)
+storageGcsPathRemoveCallback(void *const callbackData, const StorageInfo *const info)
 {
     FUNCTION_TEST_BEGIN();
         FUNCTION_TEST_PARAM_P(VOID, callbackData);
@@ -813,13 +872,12 @@ storageGcsPathRemoveCallback(void *callbackData, const StorageInfo *info)
     ASSERT(callbackData != NULL);
     ASSERT(info != NULL);
 
-    StorageGcsPathRemoveData *data = callbackData;
+    StorageGcsPathRemoveData *const data = callbackData;
 
     // Get response from prior async request
     if (data->request != NULL)
     {
-        storageGcsResponseP(data->request, .allowMissing = true);
-
+        httpResponseFree(storageGcsResponseP(data->request, .allowMissing = true));
         httpRequestFree(data->request);
         data->request = NULL;
     }
@@ -875,7 +933,7 @@ storageGcsPathRemove(THIS_VOID, const String *path, bool recurse, StorageInterfa
 
 /**********************************************************************************************************************************/
 static void
-storageGcsRemove(THIS_VOID, const String *file, StorageInterfaceRemoveParam param)
+storageGcsRemove(THIS_VOID, const String *const file, const StorageInterfaceRemoveParam param)
 {
     THIS(StorageGcs);
 
@@ -889,7 +947,7 @@ storageGcsRemove(THIS_VOID, const String *file, StorageInterfaceRemoveParam para
     ASSERT(file != NULL);
     ASSERT(!param.errorOnMissing);
 
-    storageGcsRequestP(this, HTTP_VERB_DELETE_STR, .object = file, .allowMissing = true);
+    httpResponseFree(storageGcsRequestP(this, HTTP_VERB_DELETE_STR, .object = file, .allowMissing = true));
 
     FUNCTION_LOG_RETURN_VOID();
 }
@@ -898,18 +956,19 @@ storageGcsRemove(THIS_VOID, const String *file, StorageInterfaceRemoveParam para
 static const StorageInterface storageInterfaceGcs =
 {
     .info = storageGcsInfo,
-    .infoList = storageGcsInfoList,
+    .list = storageGcsList,
     .newRead = storageGcsNewRead,
     .newWrite = storageGcsNewWrite,
     .pathRemove = storageGcsPathRemove,
     .remove = storageGcsRemove,
 };
 
-Storage *
+FN_EXTERN Storage *
 storageGcsNew(
-    const String *path, bool write, StoragePathExpressionCallback pathExpressionFunction, const String *bucket,
-    StorageGcsKeyType keyType, const String *key, size_t chunkSize, const String *endpoint, TimeMSec timeout, bool verifyPeer,
-    const String *caFile, const String *caPath)
+    const String *const path, const bool write, StoragePathExpressionCallback pathExpressionFunction, const String *const bucket,
+    const StorageGcsKeyType keyType, const String *const key, const size_t chunkSize, const KeyValue *const tag,
+    const String *const endpoint, const TimeMSec timeout, const bool verifyPeer, const String *const caFile,
+    const String *const caPath)
 {
     FUNCTION_LOG_BEGIN(logLevelDebug);
         FUNCTION_LOG_PARAM(STRING, path);
@@ -919,6 +978,7 @@ storageGcsNew(
         FUNCTION_LOG_PARAM(STRING_ID, keyType);
         FUNCTION_TEST_PARAM(STRING, key);
         FUNCTION_LOG_PARAM(SIZE, chunkSize);
+        FUNCTION_LOG_PARAM(KEY_VALUE, tag);
         FUNCTION_LOG_PARAM(STRING, endpoint);
         FUNCTION_LOG_PARAM(TIME_MSEC, timeout);
         FUNCTION_LOG_PARAM(BOOL, verifyPeer);
@@ -931,13 +991,9 @@ storageGcsNew(
     ASSERT(keyType == storageGcsKeyTypeAuto || key != NULL);
     ASSERT(chunkSize != 0);
 
-    Storage *this = NULL;
-
-    OBJ_NEW_BEGIN(StorageGcs)
+    OBJ_NEW_BEGIN(StorageGcs, .childQty = MEM_CONTEXT_QTY_MAX)
     {
-        StorageGcs *driver = OBJ_NEW_ALLOC();
-
-        *driver = (StorageGcs)
+        *this = (StorageGcs)
         {
             .interface = storageInterfaceGcs,
             .write = write,
@@ -946,17 +1002,43 @@ storageGcsNew(
             .chunkSize = chunkSize,
         };
 
+        // Create tag JSON buffer
+        if (write && tag != NULL)
+        {
+            MEM_CONTEXT_TEMP_BEGIN()
+            {
+                JsonWrite *const tagJson = jsonWriteObjectBegin(
+                    jsonWriteKeyStrId(jsonWriteObjectBegin(jsonWriteNewP()), STRID5("metadata", 0xd0240d0ad0)));
+                const StringList *const keyList = strLstSort(strLstNewVarLst(kvKeyList(tag)), sortOrderAsc);
+
+                for (unsigned int keyIdx = 0; keyIdx < strLstSize(keyList); keyIdx++)
+                {
+                    const String *const key = strLstGet(keyList, keyIdx);
+                    jsonWriteStr(jsonWriteKey(tagJson, key), varStr(kvGet(tag, VARSTR(key))));
+                }
+
+                const String *const tagStr = jsonWriteResult(jsonWriteObjectEnd(jsonWriteObjectEnd(tagJson)));
+
+                MEM_CONTEXT_PRIOR_BEGIN()
+                {
+                    this->tag = bufDup(BUFSTR(tagStr));
+                }
+                MEM_CONTEXT_PRIOR_END();
+            }
+            MEM_CONTEXT_TEMP_END();
+        }
+
         // Handle auth key types
         switch (keyType)
         {
             // Auto authentication for GCE instances
             case storageGcsKeyTypeAuto:
             {
-                driver->authUrl = httpUrlNewParseP(
+                this->authUrl = httpUrlNewParseP(
                     STRDEF("metadata.google.internal/computeMetadata/v1/instance/service-accounts/default/token"),
                     .type = httpProtocolTypeHttp);
-                driver->authClient = httpClientNew(
-                    sckClientNew(httpUrlHost(driver->authUrl), httpUrlPort(driver->authUrl), timeout, timeout), timeout);
+                this->authClient = httpClientNew(
+                    sckClientNew(httpUrlHost(this->authUrl), httpUrlPort(this->authUrl), timeout, timeout), timeout);
 
                 break;
             }
@@ -964,19 +1046,17 @@ storageGcsNew(
             // Read data from file for service keys
             case storageGcsKeyTypeService:
             {
-                KeyValue *kvKey = jsonToKv(strNewBuf(storageGetP(storageNewReadP(storagePosixNewP(FSLASH_STR), key))));
-                driver->credential = varStr(kvGet(kvKey, GCS_JSON_CLIENT_EMAIL_VAR));
-                driver->privateKey = varStr(kvGet(kvKey, GCS_JSON_PRIVATE_KEY_VAR));
+                const KeyValue *const kvKey = varKv(
+                    jsonToVar(strNewBuf(storageGetP(storageNewReadP(storagePosixNewP(FSLASH_STR), key)))));
                 const String *const uri = varStr(kvGet(kvKey, GCS_JSON_TOKEN_URI_VAR));
+                CHECK(FormatError, uri != NULL, "uri missing");
 
-                CHECK(FormatError, driver->credential != NULL && driver->privateKey != NULL && uri != NULL, "credentials missing");
-
-                driver->authUrl = httpUrlNewParseP(uri, .type = httpProtocolTypeHttps);
-
-                driver->authClient = httpClientNew(
+                this->key = strDup(key);
+                this->authUrl = httpUrlNewParseP(uri, .type = httpProtocolTypeHttps);
+                this->authClient = httpClientNew(
                     tlsClientNewP(
-                        sckClientNew(httpUrlHost(driver->authUrl), httpUrlPort(driver->authUrl), timeout, timeout),
-                        httpUrlHost(driver->authUrl), timeout, timeout, verifyPeer, .caFile = caFile, .caPath = caPath),
+                        sckClientNew(httpUrlHost(this->authUrl), httpUrlPort(this->authUrl), timeout, timeout),
+                        httpUrlHost(this->authUrl), timeout, timeout, verifyPeer, .caFile = caFile, .caPath = caPath),
                     timeout);
 
                 break;
@@ -984,33 +1064,31 @@ storageGcsNew(
 
             // Store the authentication token
             case storageGcsKeyTypeToken:
-                driver->token = strDup(key);
+                this->token = strDup(key);
                 break;
         }
 
         // Parse the endpoint to extract the host and port
         HttpUrl *url = httpUrlNewParseP(endpoint, .type = httpProtocolTypeHttps);
-        driver->endpoint = httpUrlHost(url);
+        this->endpoint = httpUrlHost(url);
 
         // Create the http client used to service requests
-        driver->httpClient = httpClientNew(
+        this->httpClient = httpClientNew(
             tlsClientNewP(
-                sckClientNew(driver->endpoint, httpUrlPort(url), timeout, timeout), driver->endpoint, timeout, timeout, verifyPeer,
+                sckClientNew(this->endpoint, httpUrlPort(url), timeout, timeout), this->endpoint, timeout, timeout, verifyPeer,
                 .caFile = caFile, .caPath = caPath),
             timeout);
 
         // Create list of redacted headers
-        driver->headerRedactList = strLstNew();
-        strLstAdd(driver->headerRedactList, HTTP_HEADER_AUTHORIZATION_STR);
-        strLstAdd(driver->headerRedactList, GCS_HEADER_UPLOAD_ID_STR);
+        this->headerRedactList = strLstNew();
+        strLstAdd(this->headerRedactList, HTTP_HEADER_AUTHORIZATION_STR);
+        strLstAdd(this->headerRedactList, GCS_HEADER_UPLOAD_ID_STR);
 
         // Create list of redacted query keys
-        driver->queryRedactList = strLstNew();
-        strLstAdd(driver->queryRedactList, GCS_QUERY_UPLOAD_ID_STR);
-
-        this = storageNew(STORAGE_GCS_TYPE, path, 0, 0, write, pathExpressionFunction, driver, driver->interface);
+        this->queryRedactList = strLstNew();
+        strLstAdd(this->queryRedactList, GCS_QUERY_UPLOAD_ID_STR);
     }
     OBJ_NEW_END();
 
-    FUNCTION_LOG_RETURN(STORAGE, this);
+    FUNCTION_LOG_RETURN(STORAGE, storageNew(STORAGE_GCS_TYPE, path, 0, 0, write, pathExpressionFunction, this, this->interface));
 }
