@@ -509,7 +509,7 @@ sub containerBuild
             {
                 $strScript .=
                     "    echo \"deb http://apt.postgresql.org/pub/repos/apt/ \$(lsb_release -s -c)-pgdg main" .
-                        ($strOS eq VM_U22 ? ' 16' : '') . "\" >> /etc/apt/sources.list.d/pgdg.list && \\\n" .
+                        "\" >> /etc/apt/sources.list.d/pgdg.list && \\\n" .
                     "    wget --quiet -O - https://www.postgresql.org/media/keys/ACCC4CF8.asc | apt-key add - && \\\n" .
                     "    apt-get update && \\\n" .
                     "    apt-get install -y --no-install-recommends postgresql-common libpq-dev && \\\n" .
@@ -617,6 +617,12 @@ sub containerBuild
                     "    echo 'PubkeyAcceptedAlgorithms=+ssh-rsa,ssh-rsa-cert-v01\@openssh.com' >> /etc/ssh/sshd_config";
             }
 
+            # Rename existing group that would conflict with our group name. This is pretty hacky but should be OK since we are the
+            # only thing running in the container.
+            $strScript .= sectionHeader() .
+                "# Rename conflicting group\n" .
+                '    sed -i s/.*\:x\:' . TEST_GROUP_ID . '\:$/' . TEST_GROUP . '\:x\:' . TEST_GROUP_ID . "\:/ /etc/group";
+
             $strScript .= sectionHeader() .
                 "# Create test user\n" .
                 '    ' . groupCreate($strOS, TEST_GROUP, TEST_GROUP_ID) . " && \\\n" .
